# What & Why preprocessing
Data preprocessing usually refers to transform the features (data) into different format which could be used in model training or model inference.

Most of the time, the training data would not be used directly. Because it contains string, date, N/A, or missing value. In order to fit the data into the model, we should preprocess the data into the numerical value and then used those for the training.

But, **we must make sure that we do the same processing for the data in the inference stage or test/evaluation stage**. We need to make sure that the data in different stages should be consistently. Otherwise, the features used for training are different from others, then the test/inference result will not be good.

# Preprocessing tips

## Date feature
Fastai implements the `add_datepart()` function, which could automatically helps to extract the feature from the date, like date of week, date of year etc.. Those additional fields expanded date-time would be useful to capture any trend/cyclical behavior.

## String feature/value
Convert those categories strings into the enum-like value. Like the fastai package function `train_cats()`, `apply_cats()`.

## Missing feature/value
Sometimes, the input data might miss some values for certain rows/columns. We need to fill those missing data before training. 

### Numerical type
There are general two solutions:
- Usually, we set the value into -1, but it depends on the scenarios.
- Create a new corresponding column (say *{column name}_na*) to indicate the missing value info. Like set the value to true if missing, otherwise false. And in the cell of original column, it will just fill with the median value of that column. Example, `fix_missing` in fastai.

### Categorical type
Pandas also automatically handles this by setting it as -1.

## Scale vars
Scale the range for each feature would be very helpful for training. 